; -------------------------------------------------------------
; This is the MASM 6+ source for a 16-bit DOS program that
; installs temporary handlers for interrupts 15h and 16h
; before executing another 16-bit DOS program. The interrupt
; 15h handler monitors for the hot key (Shift+$). When the
; hot key is detected the interrupt 16h handler switches to
; a mode where it intercepts calls (from the executed program)
; to the BIOS Get Keystroke, Check For Keystroke, Get Enhanced
; Keystroke, or Check For Enhanced Keystroke functions, and
; returns a predefined sequence of keystroke data.
;
; Although this program does pass its command line (and
; environment) to the executed program, it does not use the
; command line itself.
; -------------------------------------------------------------

    ; These parameters defined here for easy edit.
    ;
    PROGRAM_NAME      EQU "showkeys.com"
    HOTKEY_SCAN_CODE  EQU 5
    HOTKEY_SHIFT_MASK EQU 11b

    ; Declare the LoadExec structure used by the DOS Load
    ; and Execute Program function.
    ;
    LoadExec STRUCT
      leEnvironment dw 0      ; environment-block segment
      leCommandTail dd 0      ; address of command tail
      leFCB_1       dd 0      ; address of default FCB #1
      leFCB_2       dd 0      ; address of default FCB #1
    LoadExec ENDS    

    .model tiny               ; generate a COM file
    .386                      ; enable 32-bit instructions
    .stack                    ; default size (will be changed)
.code                         ; start code segmemt
.startup                      ; for tiny, expands to ORG 100h

    ; Place the data in the code segment for easy
    ; access from the interrupt handlers.

    jmp   @F                  ; jump around data

    ; Define structure var using the default initializers.
    ;
    LoadArgs      LoadExec <>

    progName      db PROGRAM_NAME,0   
    
    msgExec       db "Error executing child program, $"
    msgExit       db "press any key to exit...$"

    prevIsr15h    dd 0        ; previous Interrupt 15h handler
    prevIsr16h    dd 0        ; previous Interrupt 16h handler
    fActive       dw 0        ; set when active
    firstKey      dw 2348h    ; H
                  dw 1769h    ; i
                  dw 3920h    ;
                  dw 1454h    ; T
                  dw 2368h    ; h
                  dw 1265h    ; e
                  dw 1372h    ; r
                  dw 1265h    ; e
                  dw 0221h    ; !
    lastKey       dw 1c0dh    ; <enter>
    nextKey       dw firstKey ; offset of next key in sequence

  @@:
    ; Move stack and release unneeded memory.
    mov   bx,OFFSET theEnd
    mov   sp,bx               ; move stack
    shr   bx,4                ; convert to paragraphs
    inc   bx                  ; correct for remainder
    mov   ah,4Ah              ; set memory block size
    int   21h                 ; (ES specifies memory block)

    ; Hook Interrupt 15h.
    mov   ax,3515h            ; get interrupt vector
    int   21h                 ; returns handler address in ES:BX
    mov   WORD PTR prevIsr15h,bx
    mov   WORD PTR prevIsr15h+2,es
    mov   ax,2515h            ; set interrupt vector
    mov   dx,OFFSET Isr15h    ; DS:DX -> new interrupt handler
    int   21h

    ; Hook Interrupt 16h.
    mov   ax,3516h            ; get interrupt vector
    int   21h                 ; returns handler address in ES:BX
    mov   WORD PTR prevIsr16h,bx
    mov   WORD PTR prevIsr16h+2,es    
    mov   ax,2516h            ; set interrupt vector
    mov   dx,OFFSET Isr16h    ; DS:DX -> new interrupt handler
    int   21h

    ; Initialize LoadArgs such that child program will receive
    ; exact copy of our environment, commandTail, and FCB's.
    mov   LoadArgs.leEnvironment,0
    mov   WORD PTR LoadArgs.leCommandTail,80h
    mov   WORD PTR LoadArgs.leCommandTail+2,ds
    mov   WORD PTR LoadArgs.leFCB_1,5Ch
    mov   WORD PTR LoadArgs.leFCB_1+2,ds
    mov   WORD PTR LoadArgs.leFCB_2,6Ch
    mov   WORD PTR LoadArgs.leFCB_2+2,ds

    ; Execute program directly as child.
    mov   dx,OFFSET progName  ; DS:DX -> path
    push  ds
    pop   es
    mov   bx,OFFSET LoadArgs  ; ES:BX -> LoadExec structure
    mov   ax,4B00h            ; load and execute
    int   21h
    .if CARRY?
      mov   dx,OFFSET msgExec
      mov   ah,9              ; write string to stdout
      int   21h
    .endif

    ;Clean up and exit.
    push  ds
    mov   ax,2515h            ; set interrupt vector
    lds   dx,prevIsr15h       ; DS:DX -> interrupt handler
    int   21h                 ; restore original handler
    pop   ds
    push  ds
    mov   ax,2515h            ; set interrupt vector
    lds   dx,prevIsr16h       ; DS:DX -> interrupt handler
    int   21h                 ; restore original handler
    pop   ds

    mov   dx,OFFSET msgExit
    mov   ah,9                ; write string to stdout
    int   21h

    mov   ah,10h              ; get keystroke (will wait)
    int   16h

    .exit

; -------------------------------------------------------------
; Interrupt 15h handler. This handler tests for the hot key
; and, when it is detected, sets the active flag, absorbs the
; hot key, and returns directly to the caller (which will
; always be the BIOS as this program was the last to hook
; interrupt 15h). If the hot key is not detected this handler
; chains to the previous handler.
; -------------------------------------------------------------
CF EQU 1  ; bit value for carry flag
Isr15h proc

    push  bx
    push  es
    .if ah == 4fh             ; keyboard intercept function?
      .if al == HOTKEY_SCAN_CODE
        push  40h
        pop   es
        push  17h
        pop   bx              ; ES -> BIOS shift flags byte
        test  BYTE PTR es:[bx],HOTKEY_SHIFT_MASK
        .if !ZERO?
          mov   cs:fActive,1

          ; Clear the carry flag in the return flags (the copy
          ; of the flags that INT pushed onto the stack) so
          ; the BIOS will ignore the hot key. The IRET will
          ; pop these flags into the flags register. The
          ; contents of the stack, at this point:
          ;   SP+8  -> return flags
          ;   SP+6  -> return CS
          ;   SP+4  -> return IP
          ;   SP+2  -> preserved BX
          ;   SP    -> preserved ES
          ;
          mov   bx,sp
          xor   BYTE PTR ss:[bx+8],CF
          pop   es
          pop   bx
          iret                ; return directly to caller
        .endif
      .endif
    .endif
    pop   es
    pop   bx
    jmp   cs:prevIsr15h       ; chain to previous handler

Isr15h endp

; -------------------------------------------------------------
; Interrupt 16h handler. This handler tests the state of the
; active flag and, if it is set, returns the next key in the
; predefined keystroke sequence. When all of the keys in the
; sequence have been returned, this handler clears the active
; flag and reinitializes the sequence start address. If the
; active flag is not set this handler simply chains to the
; previous handler.
; -------------------------------------------------------------
ZF EQU 01000000b ; bit value for zero flag
Isr16h proc

    .if cs:fActive
      push  bx
      mov   bx,cs:nextKey
      .if ah == 1 || ah == 11h
        mov   ax,cs:[bx]      ; set return ascii and scan codes

        ; Clear the zero flag in the return flags (the copy of
        ; the flags that INT pushed onto the stack) to indicate
        ; that a keystroke is available. The IRET will pop
        ; these flags into the flags register. The contents of
        ; the stack, at this point:
        ;   SP+6  -> return flags
        ;   SP+4  -> return CS
        ;   SP+2  -> return IP
        ;   SP    -> preserved BX
        ;
        mov   bx,sp
        xor   BYTE PTR ss:[bx+6],ZF
        pop   bx
        iret                  ; return directly to caller
      .elseif ah == 0 || ah == 10h
        mov   ax,cs:[bx]      ; set return ascii and scan codes
        add   cs:nextKey,2    ; adjust address to next key
        .if cs:nextKey > OFFSET lastKey
          mov   cs:fActive,0
          mov   cs:nextKey,OFFSET firstKey
        .endif
        pop   bx
        iret                  ; return directly to caller
      .endif
    .endif
    jmp   cs:prevIsr16h       ; chain to previous handler

Isr16h endp

; -------------------------------------------------------------
; Allocate a new stack.
; -------------------------------------------------------------
    even                      ; specify word alignment
    dw 100h dup(0)            ; allocate 256-words
theEnd:                       ; label end of program
; -------------------------------------------------------------
end
